#!/usr/bin/env python

"""
HUMAnN2 : HMP Unified Metabolic Analysis Network 2

HUMAnN2 is a pipeline for efficiently and accurately determining
the coverage and abundance of microbial pathways in a community
from metagenomic data. Sequencing a metagenome typically produces millions
of short DNA/RNA reads.

This software is used to test the HUMAnN2 pipeline.

Dependencies: HUMAnN2 (and all HUMAnN2 dependencies)

"""

import os
import sys
import unittest

# Try to load one of the humann2 modules to check the installation
try:
    from humann2 import check
except ImportError:
    sys.exit("CRITICAL ERROR: Unable to find the HUMAnN2 python package." +
        " Please check your install.") 

# Check the python version
check.python_version()

import argparse

def parse_arguments(args):
    """ 
    Parse the arguments from the user
    """
    parser = argparse.ArgumentParser(
        description= "HUMAnN2 Test\n",
        formatter_class=argparse.RawTextHelpFormatter,
        prog="humann2_test")
    parser.add_argument(
        "--run-functional-tests-tools", 
        help="run the functional tests for tools\n", 
        action="store_true",
        default=False)
    parser.add_argument(
        "--run-functional-tests-end-to-end", 
        help="run the humann2 end to end functional tests\n", 
        action="store_true",
        default=False)
    parser.add_argument(
        "--bypass-unit-tests", 
        help="do not run the unit tests\n", 
        action="store_true",
        default=False)
    
    return parser.parse_args()

def get_testdirectory():
    """ Return the location of all of the tests """
    
    return os.path.dirname(os.path.abspath(__file__))

def get_funtionaltests_tools():
    """ Return all of the functional tests for tools"""
    
    directory_of_tests=get_testdirectory()
    
    functional_suite = unittest.TestLoader().discover(directory_of_tests,pattern='functional_tests_tools*.py')
    
    return functional_suite

def get_funtionaltests_other():
    """ Return all of the other functional tests """
    
    directory_of_tests=get_testdirectory()
    
    functional_suite = unittest.TestLoader().discover(directory_of_tests,pattern='functional_tests_humann2*.py')
    
    return functional_suite

def get_unittests():
    """ Return all of the unit tests """
    
    directory_of_tests=get_testdirectory()
    
    basic_suite = unittest.TestLoader().discover(directory_of_tests,pattern='basic_tests_*.py')
    advanced_suite = unittest.TestLoader().discover(directory_of_tests, pattern='advanced_tests_*.py')    
    
    return [basic_suite, advanced_suite]

def unittests_suite_only():
    """ Return a TestSuite of just the unit tests """
    
    return unittest.TestSuite(get_unittests())

def main():
    # Parse arguments from command line
    args=parse_arguments(sys.argv)
    
    # Get the unittests
    test_suites=[]
    if not args.bypass_unit_tests:
        test_suites=get_unittests()
    
    # Get the functional tests if requested
    if args.run_functional_tests_tools:
        test_suites+=get_funtionaltests_tools()
    if args.run_functional_tests_end_to_end:
        print("\n\nPlease note running functional end to end tests requires all dependencies of HUMAnN2.\n")
        test_suites+=get_funtionaltests_other()

    unittest.TextTestRunner(verbosity=2).run(unittest.TestSuite(test_suites))

